#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>

#include <asm/uaccess.h>

MODULE_LICENSE("Dual BSD/GPL");
MODULE_DESCRIPTION("SillyDevi module");

static char *buffer;

static int size = 10;
int silly_major = 0;  

int silly_open (struct inode *inode, struct file *filp) 
{
  printk (KERN_ALERT "SillyDev: Open \n");
  return 0;
}

int silly_release (struct inode *inode, struct file *filp) 
{
  printk (KERN_ALERT "SillyDev: Close \n");
  return 0;
}

ssize_t silly_read (struct file * filp, char * buf, size_t count, 
                    loff_t *fops) 
{  
  ssize_t ret = 0;
  
  if ((*fops) >= size) 
    return ret;
  
  if (( (*fops) + count) > size) 
    count = size - (*fops);
    
  if (copy_to_user (buf, buffer, count)) {
    ret = -1;  /* meglio sarebbe tornare il giusto errno 
                * ad esempio -EFAULT etc. etc. 
                * /usr/include/linux/errno.h */ 
  }
  else {
    ret = count;
    (*fops) += count; /* update del file position */
  }
  
  return ret;
} 

ssize_t silly_write (struct file *filp, const char *buf, size_t count, 
                     loff_t *fops) 
{  
  ssize_t ret = 0;
  
  if( (*fops) >= size) 
    return ret;
    
  if( ( (*fops) + count) > size) 	
    count = size - (*fops); 
  
  if (copy_from_user (buffer+(*fops), buf, count)) {
    ret = -1;
  }
  else {
    ret = count;
    (*fops) += count;
  }
  
  return ret;  
}

/*
 * Associazione tra operazione e funzione del driver che la implementa.
 */
struct file_operations silly_fops = {
  .open = silly_open,
  .release = silly_release,
  .read = silly_read,	
  .write = silly_write,
  .owner = THIS_MODULE /* puntatore al modulo proprietario della struttura. Necessaria
                        * affinche' il modulo non sia "uloaded" mentre les sue operazioni
                        * sono in uso. 
                        */
};

static __init int silly_init (void) {
  
  int result, i;
  
  /* per maggiori dettagli vedi link */
  buffer = kmalloc(size, GFP_KERNEL);
  if(!buffer) 
    return -ENOMEM;

  for (i=0; i<size; i++)
    buffer[i] = '\0';
  /*
   * silly_major
   * major device number se 0 il major numer viene assegnato dinamicamente 
   *       
   * "sillydev"
   * nome del device
   *         
   * silly_fops
   * file operations associate con questo device
  */    
  result = register_chrdev (silly_major, "sillydev", &silly_fops);
  if (result < 0) {
    printk(KERN_INFO "SillyDev: can't get major number\n");
    return result;
  }
  
  if (silly_major == 0)
    silly_major = result;
    
  printk (KERN_ALERT "SillyDev: is up and running with major = %d and size = %d\n", 
      silly_major, size);
 	
  return 0;
}

static __exit void silly_cleanup (void) {
  
  kfree(buffer); 
  unregister_chrdev(silly_major, "silly");
  printk (KERN_ALERT "SillyDev: bye bye\n");
  
  return;
}

module_init (silly_init);
module_exit (silly_cleanup);

module_param (size, int, S_IRUGO);
